package org.processmining.plugins.bpmnminer.causalnet;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.SwingConstants;

import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.AbstractDirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;

public class CausalGraph 
	extends AbstractDirectedGraph<CausalGraphNode, CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>> {

	protected final Map<Integer, CausalGraphNode> taskNodes;
	protected final Map<Integer, Set<CausalGraphNode>> andOutNodes;
	protected final Map<Integer, Set<CausalGraphNode>> andInNodes;
	
	protected final Set<CausalGraphEdge<CausalGraphNode, CausalGraphNode>> arcs;

	public CausalGraph() {
		super();
		getAttributeMap().put(AttributeMap.PREF_ORIENTATION, SwingConstants.WEST);
		taskNodes = new HashMap<Integer, CausalGraphNode>();
		andOutNodes = new HashMap<Integer, Set<CausalGraphNode>>();
		andInNodes = new HashMap<Integer, Set<CausalGraphNode>>();
		arcs = new LinkedHashSet<CausalGraphEdge<CausalGraphNode, CausalGraphNode>>();
	}

	// Nodes
	public synchronized CausalGraphNode addNode(int code, Set<Integer> out) {
		Set<Integer> taskSet = new HashSet<Integer>();
		taskSet.add(code);
		CausalGraphNode node = new CausalGraphNodeElement(this, code, taskSet, out);
		andOutNodes.get(code).add(node);
		graphElementAdded(node);
		addArc(getNodeByCode(code), node);
		return node;
	}
	
	public synchronized CausalGraphNode addNode(Set<Integer> in, int code) {
		Set<Integer> taskSet = new HashSet<Integer>();
		taskSet.add(code);
		CausalGraphNode node = new CausalGraphNodeElement(this, code, in, taskSet);
		andInNodes.get(code).add(node);
		graphElementAdded(node);
		addArc(node, getNodeByCode(code));
		return node;
	}
	
	public synchronized CausalGraphNode addNode(int code, String label) {
		CausalGraphNode node = new CausalGraphNodeElement(this, code, label);
		taskNodes.put(code, node);
		andOutNodes.put(code, new HashSet<CausalGraphNode>());
		andInNodes.put(code, new HashSet<CausalGraphNode>());
		graphElementAdded(node);
		return node;
	}
	
	public void removeNode(CausalGraphNode node) {
		removeSurroundingEdges(node);
		graphElementRemoved(node);
		if (!node.isAnd())
			taskNodes.remove(node.getCode());
		removeNodeFromCollection(andOutNodes.get(node.getCode()), node);
		removeNodeFromCollection(andInNodes.get(node.getCode()), node);
	}
	
	public synchronized Set<CausalGraphNode> getNodes() {
		Set<CausalGraphNode> newNodes = new HashSet<CausalGraphNode>();
		newNodes.addAll(getTaskNodes());
		newNodes.addAll(getAndNodes());
		return newNodes;
	}

	public CausalGraphNode getNodeByCode(int code) {
		return taskNodes.get(code);
	}

	public Set<CausalGraphNode> getOutAndNodesByCode(int code) {
		return andOutNodes.get(code);
	}
	
	public Set<CausalGraphNode> getInAndNodesByCode(int code) {
		return andInNodes.get(code);
	}
	
	public Set<CausalGraphNode> getTaskNodes() {
		Set<CausalGraphNode> newNodes = new HashSet<CausalGraphNode>();
		newNodes.addAll(taskNodes.values());
		return newNodes;
	}
	
	public Set<CausalGraphNode> getAndNodes() {
		Set<CausalGraphNode> newNodes = new HashSet<CausalGraphNode>();
		for (Set<CausalGraphNode> s : andOutNodes.values())
			newNodes.addAll(s);
		for (Set<CausalGraphNode> s : andInNodes.values())
			newNodes.addAll(s);
		return newNodes;
	}

	public void removeNode(DirectedGraphNode cell) {
		if (cell instanceof CausalGraphNode) {
			removeNode((CausalGraphNode) cell);
		}
	}
	
	// Edges
	public CausalGraphEdge<CausalGraphNode, CausalGraphNode> 
	addArc(CausalGraphNode source, CausalGraphNode target) {
		return addArc(source, target, null);
	}

	public synchronized CausalGraphEdge<CausalGraphNode, CausalGraphNode> 
	addArc(CausalGraphNode source, CausalGraphNode target, String label) {
		checkAddEdge(source, target);
		CausalGraphEdgeElement a = new CausalGraphEdgeElement(source, target);
		if (arcs.add(a)) {
			graphElementAdded(a);
			return a;
		} else {
			for (CausalGraphEdge<CausalGraphNode, CausalGraphNode> existing : arcs) {
				if (existing.equals(a)) {
					if (label != null) {
						existing.getAttributeMap().put(AttributeMap.LABEL, label);
					}
					return existing;
				}
			}
		}
		assert (false);
		return null;
	}

	public synchronized CausalGraphEdge<CausalGraphNode, CausalGraphNode> 
	removeArc(CausalGraphNode source, CausalGraphNode target) {
		return removeFromEdges(source, target, arcs);
	}

	public CausalGraphEdge<CausalGraphNode, CausalGraphNode> 
	removeArc(CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode> edge) {
		return removeFromEdges(edge.getSource(), edge.getTarget(), arcs);
	}

	@SuppressWarnings({"rawtypes"})
	public void removeEdge(DirectedGraphEdge edge) {
		if (edge instanceof CausalGraphEdgeElement) {
			arcs.remove(edge);
		} else {
			assert (false);
		}
		graphElementRemoved(edge);
	}

	public synchronized CausalGraphEdge<CausalGraphNode, CausalGraphNode> 
	getArc(CausalGraphNode source, CausalGraphNode target) {
		Collection<CausalGraphEdge<CausalGraphNode, CausalGraphNode>> set = 
				getEdges(source, target, arcs);
		return (set.isEmpty() ? null : set.iterator().next());
	}

	public synchronized Set<CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>> getEdges() {
		Set<CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>> edges = 
				new HashSet<CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>>();
		edges.addAll(arcs);
		return edges;
	}

	protected AbstractDirectedGraph<CausalGraphNode, CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>> 
	getEmptyClone() {
		return new CausalGraph();
	}

	protected Map<? extends DirectedGraphElement, ? extends DirectedGraphElement> 
	cloneFrom(
			DirectedGraph<CausalGraphNode, CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>> graph) {
		Map<DirectedGraphElement, DirectedGraphElement> mapping = 
				new HashMap<DirectedGraphElement, DirectedGraphElement>();

		for (CausalGraphNode a : graph.getNodes()) {
			mapping.put(a, addNode(a.getCode(), a.getLabel()));
		}

		getAttributeMap().clear();
		AttributeMap map = graph.getAttributeMap();
		for (String key : map.keySet()) {
			getAttributeMap().put(key, map.get(key));
		}
		return mapping;
	}
	
	

}
