package org.processmining.plugins.bpmnminer.converter;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramFactory;
import org.processmining.models.graphbased.directed.bpmn.BPMNEdge;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.bpmn.elements.Event;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventTrigger;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventType;
import org.processmining.models.graphbased.directed.bpmn.elements.Gateway;
import org.processmining.plugins.bpmnminer.causalnet.CausalNet;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyNet;

@Plugin(name = "Convert Causal Net to BPMN Diagram", parameterLabels = { "Causal Net" }, returnLabels = { "BPMN Diagram" }, returnTypes = { BPMNDiagram.class }, userAccessible = true, help = "Converts Causal Net into BPMN Diagram")
public class CausalNetToBPMNDiagram {
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Seppe K.L.M. vanden Broucke", email = "seppe.vandenbroucke@econ.kuleuven.be", website = "http://econ.kuleuven.be")
	@PluginVariant(variantLabel = "Default Settings", requiredParameterLabels = { 0 })
	public static BPMNDiagram convert(PluginContext context, CausalNet cn) {
		BPMNDiagram returns = toBPMNDiagram(cn);
		return returns;
	}

	public static BPMNDiagram toBPMNDiagram(CausalNet causalNet) {
		DependencyNet dependencyNet = CausalNetToDependencyNet.toDependencyNet(causalNet);

		Set<String> arcs = new HashSet<String>();
		Map<Integer, Set<Integer>> dependencyPlaces = new HashMap<Integer, Set<Integer>>();

		BPMNDiagram diagram = BPMNDiagramFactory.newBPMNDiagram("");
		
		// Setup events
		Event startEvent = diagram.addEvent("", EventType.START, EventTrigger.NONE, null, false, null);
		Event endEvent = diagram.addEvent("", EventType.END, EventTrigger.NONE, null, false, null);

		// Setup activities and gateways
		for (int t : causalNet.getTasks()) {
			dependencyPlaces.put(t, new HashSet<Integer>());

			Activity activity;
			Gateway gatewayXorIn;
			Gateway gatewayXorOut;
			Gateway gatewayAnd;

			// Add activity
			String label = "#" + t;
			if (causalNet.getLog() != null) {
				label = causalNet.getLog().getLabel(t);
			} else if (causalNet.getLabel(t) != null) {
				label = causalNet.getLabel(t);
			}
			label = "<html><span style='font-size: 70%;'>" + label + "</span></html>";
			
			activity = diagram.addActivity(label, dependencyNet.isL1Loop(t), false, false, false, false);
			activity.getAttributeMap().put("_code", "" + t);
			
			// Add xor in with ands
			if (causalNet.getInputSets(t).size() > 0) {
				gatewayXorIn = diagram.addGateway("xorIn" + t, Gateway.GatewayType.DATABASED);
				diagram.addFlow(gatewayXorIn, activity, "");
				gatewayXorIn.getAttributeMap().put("_type", "xorin");
				gatewayXorIn.getAttributeMap().put("_code", "" + t);

				for (int andCounter = 0; andCounter < causalNet.getInputSets(t).size(); andCounter++) {
					if (dependencyNet.isL1Loop(t) && causalNet.getInputSets(t).get(andCounter).size() == 1
							&& causalNet.getInputSets(t).get(andCounter).contains(t))
						continue;
					gatewayAnd = diagram.addGateway("andIn" + t + "_" + andCounter, Gateway.GatewayType.PARALLEL);
					diagram.addFlow(gatewayAnd, gatewayXorIn, "");
					gatewayAnd.getAttributeMap().put("_type", "andin");
					gatewayAnd.getAttributeMap().put("_index", "" + andCounter);
					gatewayAnd.getAttributeMap().put("_code", "" + t);
				}
			} else {
				diagram.addFlow(startEvent, activity, "");
			}

			// Add xor out with ands
			if (causalNet.getOutputSets(t).size() > 0) {
				gatewayXorOut = diagram.addGateway("xorOut" + t, Gateway.GatewayType.DATABASED);
				diagram.addFlow(activity, gatewayXorOut, "");
				gatewayXorOut.getAttributeMap().put("_type", "xorout");
				gatewayXorOut.getAttributeMap().put("_code", "" + t);

				for (int andCounter = 0; andCounter < causalNet.getOutputSets(t).size(); andCounter++) {
					if (dependencyNet.isL1Loop(t) && causalNet.getOutputSets(t).get(andCounter).size() == 1
							&& causalNet.getOutputSets(t).get(andCounter).contains(t))
						continue;
					gatewayAnd = diagram.addGateway("andOut" + t + "_" + andCounter, Gateway.GatewayType.PARALLEL);
					diagram.addFlow(gatewayXorOut, gatewayAnd, "");
					gatewayAnd.getAttributeMap().put("_type", "andout");
					gatewayAnd.getAttributeMap().put("_index", "" + andCounter);
					gatewayAnd.getAttributeMap().put("_code", "" + t);
				}
			} else {
				diagram.addFlow(activity, endEvent, "");
			}
		}
		
		// Setup dependency gateways
		for (int t : causalNet.getTasks()) {
			for (int s : causalNet.getInputTasks(t)) {
				if (t == s)
					continue;
				if (!dependencyPlaces.get(s).contains(t)) {
					dependencyPlaces.get(s).add(t);
					Gateway gatewayDep = diagram.addGateway(s + "_" + t, Gateway.GatewayType.DATABASED);
					gatewayDep.getAttributeMap().put("_type", "xordep");
					gatewayDep.getAttributeMap().put("_from", "" + s);
					gatewayDep.getAttributeMap().put("_to", "" + t);
				}
			}
			for (int u : causalNet.getOutputTasks(t)) {
				if (t == u)
					continue;
				if (!dependencyPlaces.get(t).contains(u)) {
					dependencyPlaces.get(t).add(u);
					Gateway gatewayDep = diagram.addGateway(t + "_" + u, Gateway.GatewayType.DATABASED);
					gatewayDep.getAttributeMap().put("_type", "xordep");
					gatewayDep.getAttributeMap().put("_from", "" + t);
					gatewayDep.getAttributeMap().put("_to", "" + u);
				}
			}

			// Setup flows
			for (Gateway g : diagram.getGateways()) {
				if (!g.getAttributeMap().get("_type").equals("xordep"))
					continue;
				int from = Integer.parseInt(g.getAttributeMap().get("_from").toString());
				int to = Integer.parseInt(g.getAttributeMap().get("_to").toString());
				// Link this dependency place with ands which have target in their output
				for (Gateway a : diagram.getGateways()) {
					if (a.getAttributeMap().get("_type").equals("andout")) {
						// Connect this outgoing with dependency if from equals code and to in subset
						int code = Integer.parseInt(a.getAttributeMap().get("_code").toString());
						int index = Integer.parseInt(a.getAttributeMap().get("_index").toString());
						if (code != from)
							continue;
						if (!causalNet.getOutputSets(code).get(index).contains(to))
							continue;

						if (!arcs.contains(code + index + "==>" + from + to)) {
							diagram.addFlow(a, g, "");
							arcs.add(code + index + "==>" + from + to);
						}
					}
					if (a.getAttributeMap().get("_type").equals("andin")) {
						// Connect this incoming with dependency if to equals code and from in subset
						int code = Integer.parseInt(a.getAttributeMap().get("_code").toString());
						int index = Integer.parseInt(a.getAttributeMap().get("_index").toString());
						if (code != to)
							continue;
						if (!causalNet.getInputSets(code).get(index).contains(from))
							continue;
						if (!arcs.contains(code + index + "<==" + from + to)) {
							diagram.addFlow(g, a, "");
							arcs.add(code + index + "<==" + from + to);
						}
					}
				}
			}
		}

		if (diagram.getOutEdges(startEvent).size() == 0)
			diagram.removeEvent(startEvent);
		if (diagram.getInEdges(endEvent).size() == 0)
			diagram.removeEvent(endEvent);
		
		//return diagram;
		return reduce(diagram);
	}

	public static BPMNDiagram reduce(BPMNDiagram diagram) {
		// Reducing
		boolean doing = true;
		while (doing) {
			doing = false;
			
			// REMOVE ONE IN OUT ==============================================================================
			for (Gateway g : diagram.getGateways()) {
				Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> inEdges = diagram.getInEdges(g);
				Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> outEdges = diagram.getOutEdges(g);

				if (inEdges.size() != 1 || outEdges.size() != 1)
					continue;
				BPMNEdge<? extends BPMNNode, ? extends BPMNNode> inEdge = inEdges.iterator().next();
				BPMNEdge<? extends BPMNNode, ? extends BPMNNode> outEdge = outEdges.iterator().next();

				// Connect source and with my target
				diagram.addFlow(inEdge.getSource(), outEdge.getTarget(), "");
				diagram.removeEdge(inEdge);
				diagram.removeEdge(outEdge);
				diagram.removeNode(g);
				doing = true;
				break;
			}

			if (doing)
				continue;

			// REMOVE AND SAME IN ==============================================================================
			// Get all ANDs
			Set<Gateway> andGateways = new HashSet<Gateway>();
			for (Gateway g : diagram.getGateways())
				if (g.getGatewayType().equals(Gateway.GatewayType.PARALLEL))
					andGateways.add(g);
			while (true) {
				if (andGateways.size() == 0)
					break;
				// Iterate over all ANDs
				Set<Gateway> selectedAnds = new HashSet<Gateway>();
				// Pop first AND off
				Gateway andGate = andGateways.iterator().next();
				andGateways.remove(andGate);
				selectedAnds.add(andGate);
				// Find all other ANDs which match
				for (Gateway other : andGateways) {
					if (getInNodes(diagram, andGate).equals(getInNodes(diagram, other))) {
						selectedAnds.add(other);
					}
				}
				andGateways.removeAll(selectedAnds);
				// See if there are no ANDs in input or output
				boolean clearToContinue = true;
				if (selectedAnds.size() <= 1) {
					clearToContinue = false;
				}
				for (Gateway g : selectedAnds) {
					if (!clearToContinue)
						break;
					if (isNodesContainsAnd(getInNodes(diagram, g))) {
						clearToContinue = false;
					}
					if (getOutNodes(diagram, g).size() > 1) {
						clearToContinue = false;
					}
				}
				// Perform relink
				if (clearToContinue) {
					// Remove all and nodes except one
					Gateway andToKeep = selectedAnds.iterator().next();
					Set<BPMNNode> outNodes = new HashSet<BPMNNode>();
					for (Gateway g : selectedAnds)
						outNodes.addAll(getOutNodes(diagram, g));
					selectedAnds.remove(andToKeep);
					for (Gateway g : selectedAnds)
						diagram.removeNode(g);
	
					// Destroy outgoing edges from kept 
					for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : diagram.getOutEdges(andToKeep))
						diagram.removeEdge(e);
	
					// Make new XOR and connect 
					Gateway newXor = diagram.addGateway("", Gateway.GatewayType.DATABASED);
					diagram.addFlow(andToKeep, newXor, "");
	
					// Connect new XOR to all old outgoing 
					for (BPMNNode n : outNodes)
						diagram.addFlow(newXor, n, "");
					doing = true;
					break;
				}
			}

			if (doing)
				continue;

			// REMOVE AND SAME OUT ==============================================================================
			// Get all ANDs
			andGateways = new HashSet<Gateway>();
			for (Gateway g : diagram.getGateways())
				if (g.getGatewayType().equals(Gateway.GatewayType.PARALLEL))
					andGateways.add(g);
			// Iterate over all ANDs
			while (true) {
				if (andGateways.size() == 0)
					break;
				Set<Gateway> selectedAnds = new HashSet<Gateway>();
				// Pop first AND off
				Gateway andGate = andGateways.iterator().next();
				andGateways.remove(andGate);
				selectedAnds.add(andGate);
				// Find all other ANDs which match
				for (Gateway other : andGateways) {
					if (getOutNodes(diagram, andGate).equals(getOutNodes(diagram, other))) {
						selectedAnds.add(other);
					}
				}
				andGateways.removeAll(selectedAnds);
				// See if there are no ANDs in input or output
				boolean clearToContinue = true;
				if (selectedAnds.size() <= 1) {
					clearToContinue = false;
				}
				for (Gateway g : selectedAnds) {
					if (!clearToContinue)
						break;
					if (isNodesContainsAnd(getOutNodes(diagram, g))) {
						clearToContinue = false;
					}
					if (getInNodes(diagram, g).size() > 1) {
						clearToContinue = false;
					}
				}
				// Perform relink
				if (clearToContinue) {
					// Remove all and nodes except one
					Gateway andToKeep = selectedAnds.iterator().next();
					Set<BPMNNode> inNodes = new HashSet<BPMNNode>();
					for (Gateway g : selectedAnds)
						inNodes.addAll(getInNodes(diagram, g));
					selectedAnds.remove(andToKeep);
					for (Gateway g : selectedAnds)
						diagram.removeNode(g);

					// Destroy incoming edges from kept 
					for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : diagram.getInEdges(andToKeep))
						diagram.removeEdge(e);

					// Make new XOR and connect 
					Gateway newXor = diagram.addGateway("", Gateway.GatewayType.DATABASED);
					diagram.addFlow(newXor, andToKeep, "");

					// Connect new XOR 
					for (BPMNNode n : inNodes)
						diagram.addFlow(n, newXor, "");
					doing = true;
					break;
				}
			}
			
			if (doing)
				continue;

		} // End while

		return diagram;
	}

	private static Set<BPMNNode> getInNodes(BPMNDiagram diagram, BPMNNode node) {
		Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> E = diagram.getInEdges(node);

		Set<BPMNNode> N = new HashSet<BPMNNode>();

		for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : E) {
			N.add(e.getSource());
		}

		return N;
	}

	private static Set<BPMNNode> getOutNodes(BPMNDiagram diagram, BPMNNode node) {
		Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> E = diagram.getOutEdges(node);

		Set<BPMNNode> N = new HashSet<BPMNNode>();

		for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : E) {
			N.add(e.getTarget());
		}

		return N;
	}

	private static boolean isNodesContainsAnd(Set<BPMNNode> nodes) {
		for (BPMNNode n : nodes) {
			if (n instanceof Gateway && ((Gateway) n).getGatewayType().equals(Gateway.GatewayType.PARALLEL))
				return true;
		}
		return false;
	}
}