package org.processmining.plugins.bpmnminer.converter;

import java.util.Collection;
import java.util.Set;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.bpmnminer.causalnet.CausalGraph;
import org.processmining.plugins.bpmnminer.causalnet.CausalGraphEdge;
import org.processmining.plugins.bpmnminer.causalnet.CausalGraphNode;
import org.processmining.plugins.bpmnminer.causalnet.CausalNet;

@Plugin(name = "Convert Causal Net to Causal Graph", 
		parameterLabels = { "Causal Net" }, 
		returnLabels = {"Causal Graph" }, 
		returnTypes = { CausalGraph.class }, 
		userAccessible = true, 
		help = "Converts Causal Net into Causal Graph")

public class CausalNetToCausalGraph {
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, 
			author = "Seppe K.L.M. vanden Broucke", 
			email = "seppe.vandenbroucke@econ.kuleuven.be", 
			website = "http://econ.kuleuven.be")
	@PluginVariant(variantLabel = "Default Settings", requiredParameterLabels = { 0 })
	
	public static CausalGraph convert(PluginContext context, CausalNet cn) {
		CausalGraph returns = toCausalGraph(cn);
		return returns;
	}

	public static CausalGraph toCausalGraph(CausalNet causalNet) {
		CausalGraph graph = new CausalGraph();
		for (int t : causalNet.getTasks()) {
			String label = "#" + t;
			if (causalNet.getLog() != null) {
				label = causalNet.getLog().getLabel(t);
			} else if (causalNet.getLabel(t) != null) {
				label = causalNet.getLabel(t);
			}
			
			graph.addNode(t, label);
		}
		for (int t : causalNet.getTasks()) {
			for (Set<Integer> outSet : causalNet.getOutputSets(t))
				graph.addNode(t, outSet);
			for (Set<Integer> inSet : causalNet.getInputSets(t))
				graph.addNode(inSet, t);
		}
		for (int t : causalNet.getTasks()) {
			for (CausalGraphNode andNode : graph.getOutAndNodesByCode(t)) {
				for (int o : andNode.getOutgoing()) {
					// Link this outgoing AND with the incoming ANDs for the target which contains t
					for (CausalGraphNode iandNode : graph.getInAndNodesByCode(o)) {
						if (iandNode.getIncoming().contains(t))
							graph.addArc(andNode, iandNode);
					}
				}
			}
			for (CausalGraphNode andNode : graph.getInAndNodesByCode(t)) {
				for (int i : andNode.getIncoming()) {
					// Link this incoming AND with the outgoing ANDs for the source which contains t
					for (CausalGraphNode oandNode : graph.getOutAndNodesByCode(i)) {
						if (oandNode.getOutgoing().contains(t))
							graph.addArc(oandNode, andNode);
					}
				}
			}
		}
		
		// Reducing
		boolean doing = true;
		while (doing) {
			doing = false;
			for (CausalGraphNode aNode : graph.getAndNodes()) {
				Collection<CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>> 
					inEdges = graph.getInEdges(aNode);
				Collection<CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode>> 
					outEdges = graph.getOutEdges(aNode);
				if (inEdges.size() != 1 || outEdges.size() != 1)
					continue;
				CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode> 
					inEdge = inEdges.iterator().next();
				CausalGraphEdge<? extends CausalGraphNode, ? extends CausalGraphNode> 
					outEdge = outEdges.iterator().next();
				if ((inEdge.getSource().isAnd() && !outEdge.getTarget().isAnd())
					|| (!inEdge.getSource().isAnd() && outEdge.getTarget().isAnd())) {
					// Connect source and with my target
					graph.addArc(inEdge.getSource(), outEdge.getTarget());
					graph.removeEdge(inEdge);
					graph.removeEdge(outEdge);
					graph.removeNode(aNode);
					doing = true;
					break;
				}
			}
		}
		return graph;
	}


}