package org.processmining.plugins.bpmnminer.converter;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.bpmnminer.causalnet.CausalNet;
import org.processmining.plugins.kutoolbox.utils.PetrinetUtils;
import org.processmining.plugins.reducepetrinet.PetrinetReducerPlugin;

@Plugin(name = "Convert Causal Net to Reduced Petri Net", 
		parameterLabels = { "Causal Net" }, 
		returnLabels = {"Petri net", "Marking" }, 
		returnTypes = { Petrinet.class, Marking.class }, 
		userAccessible = true, 
		help = "Converts Causal Net into Reduced Petri net")

public class CausalNetToPetrinet {
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, 
			author = "Seppe K.L.M. vanden Broucke", 
			email = "seppe.vandenbroucke@econ.kuleuven.be", 
			website = "http://econ.kuleuven.be")
	@PluginVariant(variantLabel = "Default Settings", requiredParameterLabels = { 0 })
	
	public static Object[] convert(PluginContext context, CausalNet cn) {
		Object[] objects = toPetrinet(cn);
		
		Petrinet net = (Petrinet) objects[0];
		Marking m = (Marking) objects[1];
		
		if (context != null)
			context.addConnection(new InitialMarkingConnection(net, m));

		return objects;
	}

	public static Object[] toPetrinet(CausalNet causalNet) {
		Petrinet net = PetrinetFactory.newPetrinet("");
		Map<Integer, Set<Integer>> dependencyPlaces = new HashMap<Integer, Set<Integer>>();
	//	Map<String, Set<Transition>> codeToTransitionsAct = new HashMap<String, Set<Transition>>();
	//	Map<String, Set<Transition>> codeToTransitionsTin = new HashMap<String, Set<Transition>>();
	//	Map<String, Set<Transition>> codeToTransitionsTout = new HashMap<String, Set<Transition>>();
	//	Set<Place> depPlaces = new HashSet<Place>();
		
		// Setup activities and places
		for (int t : causalNet.getTasks()) {
			dependencyPlaces.put(t, new HashSet<Integer>());
			
			Transition activity;
			Place pIn, pOut;
			Transition silent;

			// Add activity
			String label = "#" + t;
			if (causalNet.getLog() != null) {
				label = causalNet.getLog().getLabel(t);
			} else if (causalNet.getLabel(t) != null) {
				label = causalNet.getLabel(t);
			}
			activity = net.addTransition(label);
			activity.getAttributeMap().put("_code", "" + t);
			activity.getAttributeMap().put("_type", "activity");
			Set<Transition> newSet = new HashSet<>();
			newSet.add(activity);
	//		codeToTransitionsAct.put("" + t, newSet);

			// Add Place in with ands
			pIn = net.addPlace("pIn" + t);
			net.addArc(pIn, activity);
			pIn.getAttributeMap().put("_type", "pin");
			pIn.getAttributeMap().put("_code", "" + t);

			for (int andCounter = 0; andCounter < causalNet.getInputSets(t).size(); andCounter++) {
				silent = net.addTransition("");
				silent.setInvisible(true);
				net.addArc(silent, pIn);
				silent.getAttributeMap().put("_type", "tin");
				silent.getAttributeMap().put("_index", "" + andCounter);
				silent.getAttributeMap().put("_code", "" + t);
	//			if (!codeToTransitionsTin.containsKey(""+t))
	//				codeToTransitionsTin.put(""+t, new HashSet<Transition>());
	//			codeToTransitionsTin.get("" + t).add(silent);
			}

			// Add Place out with ands
			pOut = net.addPlace("pOut" + t);
			net.addArc(activity, pOut);
			pOut.getAttributeMap().put("_type", "pout");
			pOut.getAttributeMap().put("_code", "" + t);

			for (int andCounter = 0; andCounter < causalNet.getOutputSets(t).size(); andCounter++) {
				silent = net.addTransition("");
				silent.setInvisible(true);
				net.addArc(pOut, silent);
				silent.getAttributeMap().put("_type", "tout");
				silent.getAttributeMap().put("_index", "" + andCounter);
				silent.getAttributeMap().put("_code", "" + t);
	//			if (!codeToTransitionsTout.containsKey(""+t))
	//				codeToTransitionsTout.put(""+t, new HashSet<Transition>());
	//			codeToTransitionsTout.get("" + t).add(silent);
			}
			
		}

		// Add dependency places
		for (int t : causalNet.getTasks()) {
			Place pDep;
			for (int s : causalNet.getInputTasks(t)) {
				if (!dependencyPlaces.get(s).contains(t)) {
					dependencyPlaces.get(s).add(t);
					pDep = net.addPlace(s+"_"+t);
					pDep.getAttributeMap().put("_type", "pdep");
					pDep.getAttributeMap().put("_from", "" + s);
					pDep.getAttributeMap().put("_to", "" + t);
	//				depPlaces.add(pDep);
				}
			}
			for (int u : causalNet.getOutputTasks(t)) {
				if (!dependencyPlaces.get(t).contains(u)) {
					dependencyPlaces.get(t).add(u);
					pDep = net.addPlace(t+"_"+u);
					pDep.getAttributeMap().put("_type", "pdep");
					pDep.getAttributeMap().put("_from", "" + t);
					pDep.getAttributeMap().put("_to", "" + u);
	//				depPlaces.add(pDep);
				}
			}
		}
		
		// Setup flows
		for (Place d : net.getPlaces()) {
			if (!d.getAttributeMap().get("_type").equals("pdep"))
				continue;
			int from = Integer.parseInt(d.getAttributeMap().get("_from").toString());
			int to = Integer.parseInt(d.getAttributeMap().get("_to").toString());
			// Link this dependency place with silents which have target in their output
			for (Transition s : net.getTransitions()) {
				int code = Integer.parseInt(s.getAttributeMap().get("_code").toString());
				if (s.getAttributeMap().get("_type").equals("tout") &&
					s.getAttributeMap().get("_code").equals(from+"")) {
					int index = Integer.parseInt(s.getAttributeMap().get("_index").toString());
					if (!causalNet.getOutputSets(code).get(index).contains(to)) continue;
					if (net.getArc(s, d) == null) net.addArc(s, d);
				} else if (s.getAttributeMap().get("_type").equals("tin") &&
						   s.getAttributeMap().get("_code").equals(to+"")) {
					int index = Integer.parseInt(s.getAttributeMap().get("_index").toString());
					if (!causalNet.getInputSets(code).get(index).contains(from)) continue;
					if (net.getArc(d, s) == null) net.addArc(d, s);
				}
			}
		}
		
		Marking m = PetrinetUtils.getInitialMarking(net);
		Object[] result = null;
		try {
			result = PetrinetReducerPlugin.reduce(null, net, m);
		} catch (ConnectionCannotBeObtained e) {
			e.printStackTrace();
		}
		
		return result;
	}

}