package org.processmining.plugins.bpmnminer.converter;

import java.awt.Color;
import java.util.HashMap;
import java.util.Map;

import javax.swing.SwingConstants;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyGraph;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyGraphEdge;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyGraphNode;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyNet;
import org.processmining.plugins.bpmnminer.types.DFTable;
import org.processmining.plugins.bpmnminer.types.IntegerEventLog;

@Plugin(name = "Convert Dependency Net to Dependency Graph", 
		parameterLabels = { "Dependency Net" }, 
		returnLabels = {"Dependency Graph" }, 
		returnTypes = { DependencyGraph.class }, 
		userAccessible = true, 
		help = "Converts Dependency Net into Dependency Graph")

public class DependencyNetToDependencyGraph {
	
	static private float maxSize = 8;
	static private float minSize = 2;
	static private int maxColor = 200;
	static private int minColor = 30;
	static private int maxActColor = 79;
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, 
			author = "Seppe K.L.M. vanden Broucke", 
			email = "seppe.vandenbroucke@econ.kuleuven.be", 
			website = "http://econ.kuleuven.be")
	@PluginVariant(variantLabel = "Default Settings", requiredParameterLabels = { 0 })
	
	public static DependencyGraph convert(PluginContext context, DependencyNet dn) {
		DependencyGraph returns = toBasicDependencyGraph(dn);
		return returns;
	}

	@SuppressWarnings("unused")
	public static DependencyGraph toDependencyGraph(DependencyNet dependencyNet) {
		DependencyGraph graph = new DependencyGraph();
		IntegerEventLog ieLog = dependencyNet.getLog();
		DFTable dfTable = (ieLog == null) ? null : new DFTable(ieLog);
		
		int maxActCount = 0;
		
		for (int t : dependencyNet.getTasks()) {
			if (ieLog == null) 
				break;
			maxActCount = Math.max(maxActCount, dfTable.getAMeasure(t));
		}
		
		for (int t : dependencyNet.getTasks()) {
			String label = (ieLog == null) 
					? "#" + t 
					: ieLog.getLabel(t) + "<br />\r\n"
						+ "<span style='color:white;'> (" + t + ")</span>" + "<br />\r\n" 
						+ dfTable.getAMeasure(t);
			if (dependencyNet.getLabel(t) != null)
				label += "<br />\r\n" + dependencyNet.getLabel(t);
			label = "<center>" + label + "</center>";
			DependencyGraphNode node = graph.addNode(t, label);
			if (ieLog != null) {
				double r = (((double)maxActCount - (double)dfTable.getAMeasure(t)) / (double)maxActCount);
				double c = (double)maxActColor + (255D-(double)maxActColor) * r	;
				node.getAttributeMap().put(AttributeMap.FILLCOLOR, new Color(255, (int)c, (int)c));
			}
		}
		
		int maxTotalInputCount = 0;
		int maxTotalOutputCount = 0;
		int totalInputCount = 0;
		int totalOutputCount = 0;
		Map<Integer, Integer> taskInputCount = new HashMap<Integer, Integer>();
		Map<Integer, Integer> taskOutputCount = new HashMap<Integer, Integer>();
		Map<Integer, Integer> maxTaskInputCount = new HashMap<Integer, Integer>();
		Map<Integer, Integer> maxTaskOutputCount = new HashMap<Integer, Integer>();
		
		for (int t : dependencyNet.getTasks()) {
			taskInputCount.put(t, 0);
			taskOutputCount.put(t, 0);
			maxTaskInputCount.put(t, 0);
			maxTaskOutputCount.put(t, 0);
			for (int s : dependencyNet.getTaskInputs(t)) {
				int addIn = (ieLog == null) 
						? 1 
						: dfTable.getABMeasure(s, t);
				taskInputCount.put(t, taskInputCount.get(t) + addIn);
				if (addIn > maxTaskInputCount.get(t))
					maxTaskInputCount.put(t, addIn);
				if (addIn > maxTotalInputCount)
					maxTotalInputCount = addIn;
				totalInputCount += addIn;
				graph.addArc(graph.getNodeByCode(s), graph.getNodeByCode(t));
			}
			for (int u : dependencyNet.getTaskOutputs(t)) {
				int addOut = (ieLog == null) 
						? 1 
						: dfTable.getABMeasure(t, u);
				taskOutputCount.put(t, taskOutputCount.get(t) + addOut);
				if (addOut > maxTaskOutputCount.get(t))
					maxTaskOutputCount.put(t, addOut);
				if (addOut > maxTotalOutputCount)
					maxTotalOutputCount = addOut;
				totalOutputCount += addOut;
				graph.addArc(graph.getNodeByCode(t), graph.getNodeByCode(u));
			}
		}
		
		for (int t : dependencyNet.getTasks()) {
			if (ieLog == null) 
				break;
			for (int u : dependencyNet.getTaskOutputs(t)) {
				DependencyGraphEdge<DependencyGraphNode, DependencyGraphNode> arc =
						graph.getArc(graph.getNodeByCode(t), graph.getNodeByCode(u));
				
				float size = minSize + 
						((float)dfTable.getABMeasure(t, u) / (float)maxTotalOutputCount) 
						* (maxSize - minSize);
				int color = (int) (maxColor - 
						((float)dfTable.getABMeasure(t, u) / (float)maxTaskOutputCount.get(t)) 
						* (maxColor - minColor));
				
				arc.getAttributeMap().put(AttributeMap.LINEWIDTH, size);
				arc.getAttributeMap().put(AttributeMap.EDGECOLOR, new Color(color, color, color));
				arc.getAttributeMap().put(AttributeMap.SHOWLABEL, true);
				arc.getAttributeMap().put(AttributeMap.LABEL, ""+dfTable.getABMeasure(t, u));
			}
		}
		
		return graph;
	}
	
	@SuppressWarnings("unused")
	public static DependencyGraph toBasicDependencyGraph(DependencyNet dependencyNet) {
		DependencyGraph graph = new DependencyGraph();
		graph.getAttributeMap().put(AttributeMap.PREF_ORIENTATION, SwingConstants.WEST);
		
		IntegerEventLog ieLog = dependencyNet.getLog();
		DFTable dfTable = (ieLog == null) ? null : new DFTable(ieLog);
		
		for (int t : dependencyNet.getTasks()) {
			String label = (ieLog == null) 
					? "#" + t 
					: ieLog.getLabel(t) + "\n"
						+ dfTable.getAMeasure(t);
			if (dependencyNet.getLabel(t) != null)
				label += "\n" + dependencyNet.getLabel(t);
			DependencyGraphNode node = graph.addNode(t, label);
		}
		
		int maxTotalInputCount = 0;
		int maxTotalOutputCount = 0;
		int totalInputCount = 0;
		int totalOutputCount = 0;
		Map<Integer, Integer> taskInputCount = new HashMap<Integer, Integer>();
		Map<Integer, Integer> taskOutputCount = new HashMap<Integer, Integer>();
		Map<Integer, Integer> maxTaskInputCount = new HashMap<Integer, Integer>();
		Map<Integer, Integer> maxTaskOutputCount = new HashMap<Integer, Integer>();
		
		for (int t : dependencyNet.getTasks()) {
			taskInputCount.put(t, 0);
			taskOutputCount.put(t, 0);
			maxTaskInputCount.put(t, 0);
			maxTaskOutputCount.put(t, 0);
			for (int s : dependencyNet.getTaskInputs(t)) {
				int addIn = (ieLog == null) 
						? 1 
						: dfTable.getABMeasure(s, t);
				taskInputCount.put(t, taskInputCount.get(t) + addIn);
				if (addIn > maxTaskInputCount.get(t))
					maxTaskInputCount.put(t, addIn);
				if (addIn > maxTotalInputCount)
					maxTotalInputCount = addIn;
				totalInputCount += addIn;
				graph.addArc(graph.getNodeByCode(s), graph.getNodeByCode(t));
			}
			for (int u : dependencyNet.getTaskOutputs(t)) {
				int addOut = (ieLog == null) 
						? 1 
						: dfTable.getABMeasure(t, u);
				taskOutputCount.put(t, taskOutputCount.get(t) + addOut);
				if (addOut > maxTaskOutputCount.get(t))
					maxTaskOutputCount.put(t, addOut);
				if (addOut > maxTotalOutputCount)
					maxTotalOutputCount = addOut;
				totalOutputCount += addOut;
				graph.addArc(graph.getNodeByCode(t), graph.getNodeByCode(u));
			}
		}
		
		for (int t : dependencyNet.getTasks()) {
			if (ieLog == null) 
				break;
			for (int u : dependencyNet.getTaskOutputs(t)) {
				DependencyGraphEdge<DependencyGraphNode, DependencyGraphNode> arc =
						graph.getArc(graph.getNodeByCode(t), graph.getNodeByCode(u));
				
				float size = minSize + 
						((float)dfTable.getABMeasure(t, u) / (float)maxTotalOutputCount) 
						* (maxSize - minSize);
				int color = (int) (maxColor - 
						((float)dfTable.getABMeasure(t, u) / (float)maxTaskOutputCount.get(t)) 
						* (maxColor - minColor));
				
				arc.getAttributeMap().put(AttributeMap.LINEWIDTH, size);
				arc.getAttributeMap().put(AttributeMap.EDGECOLOR, new Color(color, color, color));
				arc.getAttributeMap().put(AttributeMap.SHOWLABEL, true);
				arc.getAttributeMap().put(AttributeMap.LABEL, ""+dfTable.getABMeasure(t, u));
			}
		}
		
		return graph;
	}

	

}