package org.processmining.plugins.bpmnminer.importexport;

import java.io.BufferedInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.processmining.contexts.uitopia.annotations.UIImportPlugin;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.bpmnminer.causalnet.CausalNet;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

@Plugin(name = "Import Causal Net",
	parameterLabels = { "Filename" }, 
	returnLabels = { "CausalNet" },
	returnTypes = { CausalNet.class })
@UIImportPlugin(description = "CNET files", extensions = { "cnet" })

public class ImportCNet extends AbstractImportPlugin {

	protected FileFilter getFileFilter() {
		return new FileNameExtensionFilter("CNET files", "cnet");
	}

	protected CausalNet importFromStream(PluginContext context, InputStream input,  
			String filename, long fileSizeInBytes) {
		
		CnetXmlHandler handler = new CnetXmlHandler();
		try {
			BufferedInputStream bis = new BufferedInputStream(input);
			SAXParserFactory parserFactory = SAXParserFactory.newInstance();
			parserFactory.setNamespaceAware(false);
			SAXParser parser = parserFactory.newSAXParser();
			parser.parse(bis, handler);
			bis.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		CausalNet net = new CausalNet();
		
		for (int e : handler.getActivities()) {
			net.addTask(e);
			net.setLabel(e, handler.getEventClasses().get(e));
			
		}
		
		for (Entry<Integer, List<Set<Integer>>> e : handler.getInConnections().entrySet()) {
			for (Set<Integer> iset : e.getValue()) {
				net.addInputSet(e.getKey(), iset);
			}
		}

		for (Entry<Integer, List<Set<Integer>>> e : handler.getOutConnections().entrySet()) {
			for (Set<Integer> oset : e.getValue()) {
				net.addOutputSet(e.getKey(), oset);
			}
		}
		
		return net;
	}
	
	protected class CnetXmlHandler extends DefaultHandler {
		protected Map<Integer, String> eventClasses;
		protected Set<Integer> activities;
		protected Map<Integer, List<Set<Integer>>> inConnections;
		protected Map<Integer, List<Set<Integer>>> outConnections;
		protected int size = 0;
		
		boolean parsingActivities = false;
		boolean parsingActivity = false;
		boolean parsingConnections = false;
		boolean parsingInputSets = false;
		boolean parsingOutputSets = false;
		boolean parsingInputSet = false;
		boolean parsingOutputSet = false;
		int currentActivity = 0;
		Set<Integer> currentSubSet = new HashSet<Integer>();
		StringBuffer accumulator = new StringBuffer();
		
		public CnetXmlHandler() {
			eventClasses = new HashMap<Integer, String>();
			activities = new HashSet<Integer>();
			inConnections = new HashMap<Integer, List<Set<Integer>>>();
			outConnections = new HashMap<Integer, List<Set<Integer>>>();
		}

		@Override
		public void characters(char[] buffer, int start, int length) {
			accumulator.append(buffer, start, length);
		}
		
		@Override
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {
			accumulator.setLength(0);
			String tagName = localName.trim();
			if (tagName.length() == 0)
				tagName = qName;
			if (tagName.equalsIgnoreCase("activities")) {
				parsingActivities = true;
			}else if (tagName.equalsIgnoreCase("causalnet")) {
				size = Integer.parseInt(attributes.getValue("size"));
			}else if (tagName.equalsIgnoreCase("activity")) {
				parsingActivity = true;
				int identifier = Integer.parseInt(attributes.getValue("identifier"));
				if (parsingConnections == true && (parsingInputSets || parsingOutputSets)) {
					currentSubSet.add(identifier);
				} else {
					currentActivity = identifier;
					activities.add(identifier);
				}
			}else if (tagName.equalsIgnoreCase("label")) {
				// Accumulated in endElement
			}else if (tagName.equalsIgnoreCase("connections")) {
				parsingConnections = true;
			}else if (tagName.equalsIgnoreCase("inputsets")) {
				parsingInputSets = true;
			}else if (tagName.equalsIgnoreCase("outputsets")) {
				parsingOutputSets = true;
			}else if (tagName.equalsIgnoreCase("inputset")) {
				currentSubSet = new HashSet<Integer>();
			}else if (tagName.equalsIgnoreCase("outputset")) {
				currentSubSet = new HashSet<Integer>();
			}
		}

		@Override
		public void endElement(String uri, String localName, String qName)
				throws SAXException {
			String tagName = localName.trim();
			if (tagName.length() == 0)
				tagName = qName;
			
			if (tagName.equalsIgnoreCase("activities")) {
				parsingActivities = false;
			}else if (tagName.equalsIgnoreCase("activity")) {
				parsingActivity = false;
			}else if (tagName.equalsIgnoreCase("label")) {
				eventClasses.put(currentActivity, accumulator.toString().trim());
			}else if (tagName.equalsIgnoreCase("connections")) {
				parsingConnections = false;
			}else if (tagName.equalsIgnoreCase("inputsets")) {
				parsingInputSets = false;
			}else if (tagName.equalsIgnoreCase("outputsets")) {
				parsingOutputSets = false;
			}else if (tagName.equalsIgnoreCase("inputset")) {
				if (!inConnections.containsKey(currentActivity))
					inConnections.put(currentActivity, new ArrayList<Set<Integer>>());
				inConnections.get(currentActivity).add(currentSubSet);
			}else if (tagName.equalsIgnoreCase("outputset")) {
				if (!outConnections.containsKey(currentActivity))
					outConnections.put(currentActivity, new ArrayList<Set<Integer>>());
				outConnections.get(currentActivity).add(currentSubSet);
			}
		}

		public Map<Integer, String> getEventClasses() {
			return eventClasses;
		}

		public Set<Integer> getActivities() {
			return activities;
		}

		public Map<Integer, List<Set<Integer>>> getInConnections() {
			return inConnections;
		}

		public Map<Integer, List<Set<Integer>>> getOutConnections() {
			return outConnections;
		}

		public int getSize() {
			return size;
		}

	}

}


