package org.processmining.plugins.bpmnminer.miner;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramFactory;
import org.processmining.models.graphbased.directed.bpmn.BPMNEdge;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.bpmn.elements.Event;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventTrigger;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventType;
import org.processmining.models.graphbased.directed.bpmn.elements.Flow;
import org.processmining.models.graphbased.directed.bpmn.elements.Gateway;
import org.processmining.models.graphbased.directed.bpmn.elements.Swimlane;
import org.processmining.models.graphbased.directed.bpmn.elements.SwimlaneType;
import org.processmining.plugins.bpmnminer.causalnet.CausalNet;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyNet;
import org.processmining.plugins.bpmnminer.types.EventLogTaskMapper;
import org.processmining.plugins.bpmnminer.types.IntegerEventLog;
import org.processmining.plugins.bpmnminer.types.MinerSettings;
import org.processmining.plugins.bpmnminer.types.RelationalMetrics;

public class BPMNMiner {

	protected final XLog log;
	protected final MinerSettings settings;
	protected final IntegerEventLog ilog;
	protected final RelationalMetrics metrics;
	protected final EventLogTaskMapper mapper;
	
	private DependencyNet dependencyNet;
	private CausalNet causalNet;
	private BPMNDiagram bpmnDiagram;
	
	private boolean deriveOriginatorSwimLanes;
	
	public BPMNMiner(XLog l, MinerSettings s) {
		log = l;
		settings = s;
		mapper = new EventLogTaskMapper(log, settings.classifier);
		mapper.setup(settings.backwardContextSize, 
				settings.forwardContextSize, 
				settings.useUniqueStartEndTasks, 
				settings.collapseL1l,
				settings.duplicateThreshold);
		ilog = mapper.getIntegerLog();
		metrics = new RelationalMetrics(ilog, settings);
		
	}
	
	public BPMNDiagram getBPMNDiagram() {
		return bpmnDiagram;
	}

	public CausalNet getCausalNet() {
		return causalNet;
	}

	public DependencyNet getDependencyNet() {
		return dependencyNet;
	}

	public EventLogTaskMapper getMapper() {
		return mapper;
	}

	public BPMNDiagram mine() {
		FodinaMiner miner = new FodinaMiner(ilog, metrics, settings);
		miner.mine();
		
		this.dependencyNet = miner.getDependencyNet();
		this.causalNet = miner.getCausalNet();
		
		mineBPMNDiagram();

		return getBPMNDiagram();
	}
	
	public void mineBPMNDiagram() {
		System.out.println("FODINA: Annotating BPMN diagram");
		
		// Step 1: derive originator information
		Map<Integer, Set<String>> originators = null;
		List<Set<String>> swimLanes = null;		
		Map<Integer, Integer> taskToSwimlane = null;
		
		if (deriveOriginatorSwimLanes) {
			originators = mineTaskOriginators();
			swimLanes = mineSwimlanes(originators);		
			taskToSwimlane = mineTaskSwimlaneMapping(originators, swimLanes);
		}
		
		// Step 2: construct the bpmn diagram
		causalNetToBPMNDiagram(taskToSwimlane, swimLanes);		
		reduceBPMNDiagramFull();		
				
		System.out.println("FODINA: BPMN diagram mined");
	}
	
	public boolean isDeriveOriginatorSwimLanes() {
		return deriveOriginatorSwimLanes;
	}

	public void setDeriveOriginatorSwimLanes(boolean deriveOriginatorSwimLanes) {
		this.deriveOriginatorSwimLanes = deriveOriginatorSwimLanes;
	}

	private void reduceBPMNDiagramOneInOut() {
		boolean doing = true;
		while (doing) {
			doing = false;
			
			// REMOVE ONE IN OUT ==============================================================================
			for (Gateway g : bpmnDiagram.getGateways()) {
				Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> inEdges = bpmnDiagram.getInEdges(g);
				Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> outEdges = bpmnDiagram.getOutEdges(g);
	
				if (inEdges.size() != 1 || outEdges.size() != 1)
					continue;
				BPMNEdge<? extends BPMNNode, ? extends BPMNNode> inEdge = inEdges.iterator().next();
				BPMNEdge<? extends BPMNNode, ? extends BPMNNode> outEdge = outEdges.iterator().next();
	
				// Connect source and with my target
				bpmnDiagram.addFlow(inEdge.getSource(), outEdge.getTarget(), "");
				bpmnDiagram.removeEdge(inEdge);
				bpmnDiagram.removeEdge(outEdge);
				bpmnDiagram.removeNode(g);
				doing = true;
				break;
			}
		}
	}
	
	private void reduceBPMNDiagramSameAndOuts() {
		boolean doing = true;
		while (doing) {
			doing = false;
			
			// REMOVE AND SAME OUT ==============================================================================
			// Get all ANDs
			Set<Gateway> andGateways = new HashSet<Gateway>();
			for (Gateway g : bpmnDiagram.getGateways())
				if (g.getGatewayType().equals(Gateway.GatewayType.PARALLEL))
					andGateways.add(g);
			while (true) {
				if (andGateways.size() == 0)
					break;
				// Iterate over all ANDs
				Set<Gateway> selectedAnds = new HashSet<Gateway>();
				// Pop first AND off
				Gateway andGate = andGateways.iterator().next();
				andGateways.remove(andGate);
				selectedAnds.add(andGate);
				// Find all other ANDs which match
				for (Gateway other : andGateways)
					if (getOutNodes(bpmnDiagram, andGate).equals(getOutNodes(bpmnDiagram, other)))
						selectedAnds.add(other);
				andGateways.removeAll(selectedAnds);
				// See if there are no ANDs in input or output
				boolean clearToContinue = true;
				if (selectedAnds.size() <= 1) {
					clearToContinue = false;
				}
				for (Gateway g : selectedAnds) {
					if (!clearToContinue)
						break;
					if (isNodesContainsAnd(getOutNodes(bpmnDiagram, g)))
						clearToContinue = false;
					if (getInNodes(bpmnDiagram, g).size() > 1)
						clearToContinue = false;
				}
				// Perform relink
				if (clearToContinue) {
					// Remove all and nodes except one
					Gateway andToKeep = selectedAnds.iterator().next();
					Set<BPMNNode> inNodes = new HashSet<BPMNNode>();
					for (Gateway g : selectedAnds)
						inNodes.addAll(getInNodes(bpmnDiagram, g));
					selectedAnds.remove(andToKeep);
					for (Gateway g : selectedAnds)
						bpmnDiagram.removeNode(g);
	
					// Destroy incoming edges from kept 
					for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : bpmnDiagram.getInEdges(andToKeep))
						bpmnDiagram.removeEdge(e);
	
					// Make new XOR and connect 
					Gateway newXor = bpmnDiagram.addGateway("", Gateway.GatewayType.DATABASED, andToKeep.getParentLane());
					
					bpmnDiagram.addFlow(newXor, andToKeep, "");
	
					// Connect new XOR to all old outgoing 
					for (BPMNNode n : inNodes)
						bpmnDiagram.addFlow(n, newXor, "");
					doing = true;
					break;
				}
			}
		}
	}
	
	private void reduceBPMNDiagramSameAndIns() {
		boolean doing = true;
		while (doing) {
			doing = false;
			
			// REMOVE AND SAME IN ==============================================================================
			// Get all ANDs
			Set<Gateway> andGateways = new HashSet<Gateway>();
			for (Gateway g : bpmnDiagram.getGateways())
				if (g.getGatewayType().equals(Gateway.GatewayType.PARALLEL))
					andGateways.add(g);
			while (true) {
				if (andGateways.size() == 0)
					break;
				// Iterate over all ANDs
				Set<Gateway> selectedAnds = new HashSet<Gateway>();
				// Pop first AND off
				Gateway andGate = andGateways.iterator().next();
				andGateways.remove(andGate);
				selectedAnds.add(andGate);
				// Find all other ANDs which match
				for (Gateway other : andGateways)
					if (getInNodes(bpmnDiagram, andGate).equals(getInNodes(bpmnDiagram, other)))
						selectedAnds.add(other);
				andGateways.removeAll(selectedAnds);
				// See if there are no ANDs in input or output
				boolean clearToContinue = true;
				if (selectedAnds.size() <= 1) {
					clearToContinue = false;
				}
				for (Gateway g : selectedAnds) {
					if (!clearToContinue)
						break;
					if (isNodesContainsAnd(getInNodes(bpmnDiagram, g)))
						clearToContinue = false;
					if (getOutNodes(bpmnDiagram, g).size() > 1)
						clearToContinue = false;
				}
				// Perform relink
				if (clearToContinue) {
					// Remove all and nodes except one
					Gateway andToKeep = selectedAnds.iterator().next();
					Set<BPMNNode> outNodes = new HashSet<BPMNNode>();
					for (Gateway g : selectedAnds)
						outNodes.addAll(getOutNodes(bpmnDiagram, g));
					selectedAnds.remove(andToKeep);
					for (Gateway g : selectedAnds)
						bpmnDiagram.removeNode(g);
	
					// Destroy outgoing edges from kept 
					for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : bpmnDiagram.getOutEdges(andToKeep))
						bpmnDiagram.removeEdge(e);
	
					// Make new XOR and connect 
					Gateway newXor = bpmnDiagram.addGateway("", Gateway.GatewayType.DATABASED, andToKeep.getParentLane());
					bpmnDiagram.addFlow(andToKeep, newXor, "");
	
					// Connect new XOR to all old outgoing 
					for (BPMNNode n : outNodes)
						bpmnDiagram.addFlow(newXor, n, "");
					doing = true;
					break;
				}
			}
		}
	}
	
	private void reduceBPMNDiagramFull() {
		reduceBPMNDiagramOneInOut();
		reduceBPMNDiagramSameAndIns();
		reduceBPMNDiagramSameAndOuts();
		reduceBPMNDiagramOneInOut();
	}

	private void causalNetToBPMNDiagram(Map<Integer, Integer> taskToSwimlane, List<Set<String>> swimLanes) {
		Set<String> arcs = new HashSet<String>();
		Map<Integer, Set<Integer>> dependencyPlaces = new HashMap<Integer, Set<Integer>>();
	
		BPMNDiagram diagram = BPMNDiagramFactory.newBPMNDiagram("");
		List<Swimlane> slanes = new ArrayList<Swimlane>();
		if (deriveOriginatorSwimLanes) {
			for (int i = 0; i < swimLanes.size(); i++) {
				Swimlane lane = diagram.addSwimlane("Pool_"+(i+1), null, SwimlaneType.LANE);
				slanes.add(lane);
			}
		}
		
		
		// Setup eventsd
		Event startEvent = diagram.addEvent("", EventType.START, EventTrigger.NONE, 
				null, (!deriveOriginatorSwimLanes || settings.useUniqueStartEndTasks) ? 
						null : slanes.get(taskToSwimlane.get(dependencyNet.getStartTask())), false, null);
		Event endEvent = diagram.addEvent("", EventType.END, EventTrigger.NONE, 
				null, (!deriveOriginatorSwimLanes || settings.useUniqueStartEndTasks) ? 
						null : slanes.get(taskToSwimlane.get(dependencyNet.getEndTask())), false, null);
				
		// Setup activities and gateways
		for (int t : causalNet.getTasks()) {
			dependencyPlaces.put(t, new HashSet<Integer>());
	
			Activity activity;
			Gateway gatewayXorIn;
			Gateway gatewayXorOut;
			Gateway gatewayAnd;
	
			// Add activity
			String label =  causalNet.getLabel(t);
			Integer swimlane = (taskToSwimlane == null) ? null : taskToSwimlane.get(t);
			
			activity = diagram.addActivity(label, dependencyNet.isL1Loop(t), false, false, false, false, 
					(swimlane == null) ? null : slanes.get(swimlane));
			activity.getAttributeMap().put("_code", "" + t);
			
			// Add xor in with ands
			if (causalNet.getInputSets(t).size() > 0) {
				gatewayXorIn = diagram.addGateway("xorIn" + t, Gateway.GatewayType.DATABASED,
						(swimlane == null) ? null : slanes.get(swimlane));
				diagram.addFlow(gatewayXorIn, activity, "");
				gatewayXorIn.getAttributeMap().put("_type", "xorin");
				gatewayXorIn.getAttributeMap().put("_code", "" + t);
	
				for (int andCounter = 0; andCounter < causalNet.getInputSets(t).size(); andCounter++) {
					if (dependencyNet.isL1Loop(t) && causalNet.getInputSets(t).get(andCounter).size() == 1
							&& causalNet.getInputSets(t).get(andCounter).contains(t))
						continue;
					gatewayAnd = diagram.addGateway("andIn" + t + "_" + andCounter, Gateway.GatewayType.PARALLEL,
							(swimlane == null) ? null : slanes.get(swimlane));
					diagram.addFlow(gatewayAnd, gatewayXorIn, "");
					gatewayAnd.getAttributeMap().put("_type", "andin");
					gatewayAnd.getAttributeMap().put("_index", "" + andCounter);
					gatewayAnd.getAttributeMap().put("_code", "" + t);
				}
			} else {
				diagram.addFlow(startEvent, activity, "");
			}
	
			// Add xor out with ands
			if (causalNet.getOutputSets(t).size() > 0) {
				gatewayXorOut = diagram.addGateway("xorOut" + t, Gateway.GatewayType.DATABASED,
						(swimlane == null) ? null : slanes.get(swimlane));
				diagram.addFlow(activity, gatewayXorOut, "");
				gatewayXorOut.getAttributeMap().put("_type", "xorout");
				gatewayXorOut.getAttributeMap().put("_code", "" + t);
	
				for (int andCounter = 0; andCounter < causalNet.getOutputSets(t).size(); andCounter++) {
					if (dependencyNet.isL1Loop(t) && causalNet.getOutputSets(t).get(andCounter).size() == 1
							&& causalNet.getOutputSets(t).get(andCounter).contains(t))
						continue;
					gatewayAnd = diagram.addGateway("andOut" + t + "_" + andCounter, Gateway.GatewayType.PARALLEL,
							(swimlane == null) ? null : slanes.get(swimlane));
					diagram.addFlow(gatewayXorOut, gatewayAnd, "");
					gatewayAnd.getAttributeMap().put("_type", "andout");
					gatewayAnd.getAttributeMap().put("_index", "" + andCounter);
					gatewayAnd.getAttributeMap().put("_code", "" + t);
				}
			} else {
				diagram.addFlow(activity, endEvent, "");
			}
		}
		
		// Setup dependency gateways
		for (int t : causalNet.getTasks()) {
			Integer swimlane = (taskToSwimlane == null) ? null : taskToSwimlane.get(t);
			for (int s : causalNet.getInputTasks(t)) {
				if (t == s)
					continue;
				if (!dependencyPlaces.get(s).contains(t)) {
					dependencyPlaces.get(s).add(t);
					Gateway gatewayDep = diagram.addGateway(s + "_" + t, Gateway.GatewayType.DATABASED,
							(swimlane == null) ? null : slanes.get(swimlane));
					gatewayDep.getAttributeMap().put("_type", "xordep");
					gatewayDep.getAttributeMap().put("_from", "" + s);
					gatewayDep.getAttributeMap().put("_to", "" + t);
				}
			}
			for (int u : causalNet.getOutputTasks(t)) {
				if (t == u)
					continue;
				if (!dependencyPlaces.get(t).contains(u)) {
					dependencyPlaces.get(t).add(u);
					Gateway gatewayDep = diagram.addGateway(t + "_" + u, Gateway.GatewayType.DATABASED,
							(swimlane == null) ? null : slanes.get(swimlane));
					gatewayDep.getAttributeMap().put("_type", "xordep");
					gatewayDep.getAttributeMap().put("_from", "" + t);
					gatewayDep.getAttributeMap().put("_to", "" + u);
				}
			}
	
			// Setup flows
			for (Gateway g : diagram.getGateways()) {
				if (!g.getAttributeMap().get("_type").equals("xordep"))
					continue;
				int from = Integer.parseInt(g.getAttributeMap().get("_from").toString());
				int to = Integer.parseInt(g.getAttributeMap().get("_to").toString());
				// Link this dependency place with ands which have target in their output
				for (Gateway a : diagram.getGateways()) {
					if (a.getAttributeMap().get("_type").equals("andout")) {
						// Connect this outgoing with dependency if from equals code and to in subset
						int code = Integer.parseInt(a.getAttributeMap().get("_code").toString());
						int index = Integer.parseInt(a.getAttributeMap().get("_index").toString());
						if (code != from)
							continue;
						if (!causalNet.getOutputSets(code).get(index).contains(to))
							continue;
	
						if (!arcs.contains(code + index + "==>" + from + to)) {
							diagram.addFlow(a, g, "");
							arcs.add(code + index + "==>" + from + to);
						}
					}
					if (a.getAttributeMap().get("_type").equals("andin")) {
						// Connect this incoming with dependency if to equals code and from in subset
						int code = Integer.parseInt(a.getAttributeMap().get("_code").toString());
						int index = Integer.parseInt(a.getAttributeMap().get("_index").toString());
						if (code != to)
							continue;
						if (!causalNet.getInputSets(code).get(index).contains(from))
							continue;
						if (!arcs.contains(code + index + "<==" + from + to)) {
							diagram.addFlow(g, a, "");
							arcs.add(code + index + "<==" + from + to);
						}
					}
				}
			}
		}
	
		if (diagram.getOutEdges(startEvent).size() == 0)
			diagram.removeEvent(startEvent);
		if (diagram.getInEdges(endEvent).size() == 0)
			diagram.removeEvent(endEvent);
		
		
		bpmnDiagram = diagram;
		
		if (deriveOriginatorSwimLanes)
			fixSwimLanes();
		
	}

	private void fixSwimLanes() {
		Set<Flow> flows = new HashSet<Flow>();
		for (Flow f : bpmnDiagram.getFlows()) {
			BPMNNode s = f.getSource();
			BPMNNode t = f.getTarget();
			if (s.getParentSwimlane() == null || t.getParentSwimlane() == null)
				continue;
			if (s.getParentSwimlane().equals(t.getParentSwimlane()))
				continue;
			flows.add(f);
		}
		for (Flow f : flows) {
			bpmnDiagram.removeEdge(f);
			bpmnDiagram.addMessageFlow(f.getSource(), f.getTarget(), "");
			
		}
	}

	private Map<Integer, Integer> mineTaskSwimlaneMapping(
			Map<Integer, Set<String>> originators, List<Set<String>> swimLanes) {
		Map<Integer, Integer> mapping = new HashMap<Integer, Integer>();
		for (Integer t : originators.keySet()) {
			Set<String> a = originators.get(t);
			for (int i = 0; i < swimLanes.size(); i++) {
				Set<String> b = new HashSet<String>(swimLanes.get(i));
				b.retainAll(a);
				if (b.size() > 0) {
					mapping.put(t, i);
					break;
				}
			}
		}
		return mapping;
	}

	private List<Set<String>> mineSwimlanes(
			Map<Integer, Set<String>> originators) {
		List<Set<String>> swimlanes = new ArrayList<Set<String>>();
		Set<String> originatorSet = new HashSet<String>();
		for (Set<String> c : originators.values())
			originatorSet.addAll(c);
		
		for (String orgel : originatorSet) {
			swimlanes.add(new HashSet<String>());
			swimlanes.get(swimlanes.size()-1).add(orgel);
		}
		
		boolean changesMade = true;
		while (changesMade) {
			changesMade = false;
			for (int i = 0; i < swimlanes.size(); i++) {
				for (int j = i+1 ; j < swimlanes.size(); j++) {
					// Does there exist a task both executable by an org1 and org2
					for (int t : originators.keySet()) {
						Set<String> a = new HashSet<String>(swimlanes.get(i));
						Set<String> b = new HashSet<String>(swimlanes.get(j));
						a.retainAll(originators.get(t));
						b.retainAll(originators.get(t));
						if (a.size() > 0 && b.size() > 0) {
							swimlanes.get(i).addAll(swimlanes.get(j));
							swimlanes.remove(j);
							changesMade = true;
							break;
						}
					}
				}
			}
		}
		
		return swimlanes;
	}

	private Map<Integer, Set<String>> mineTaskOriginators() {
		Map<Integer, Set<String>> taskOriginators = new HashMap<Integer, Set<String>>();
		for (XTrace t : log) {
			for (int pos = 0; pos < t.size(); pos++) {
				Integer ti = mapper.getIntegerFragment(t, pos);
				if (ti == null) continue;
				if (!taskOriginators.containsKey(ti))
					taskOriginators.put(ti, new HashSet<String>());
				XEvent e = t.get(pos);
				String role = XOrganizationalExtension.instance().extractRole(e);
				String group = XOrganizationalExtension.instance().extractGroup(e);
				String resource = XOrganizationalExtension.instance().extractResource(e);
				String toAdd = null;
				if (role != null && settings.organizationalField.equals(XOrganizationalExtension.KEY_ROLE))
					toAdd = role;
				if (group != null && settings.organizationalField.equals(XOrganizationalExtension.KEY_GROUP))
					toAdd = group;
				if (resource != null && settings.organizationalField.equals(XOrganizationalExtension.KEY_RESOURCE))
					toAdd = resource;
				if (toAdd != null)
					taskOriginators.get(ti).add(toAdd);
			}
		}
		for (int t : taskOriginators.keySet())
			if (taskOriginators.get(t).size() == 0)
				taskOriginators.get(t).add("__default__");
		return taskOriginators;
	}

	private static Set<BPMNNode> getInNodes(BPMNDiagram diagram, BPMNNode node) {
		Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> E = diagram.getInEdges(node);

		Set<BPMNNode> N = new HashSet<BPMNNode>();

		for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : E) {
			N.add(e.getSource());
		}

		return N;
	}

	private static Set<BPMNNode> getOutNodes(BPMNDiagram diagram, BPMNNode node) {
		Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> E = diagram.getOutEdges(node);

		Set<BPMNNode> N = new HashSet<BPMNNode>();

		for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> e : E) {
			N.add(e.getTarget());
		}

		return N;
	}

	private static boolean isNodesContainsAnd(Set<BPMNNode> nodes) {
		for (BPMNNode n : nodes) {
			if (n instanceof Gateway && ((Gateway) n).getGatewayType().equals(Gateway.GatewayType.PARALLEL))
				return true;
		}
		return false;
	}
}
