package org.processmining.plugins.bpmnminer.plugins;

import java.awt.Color;
import java.util.Map;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramFactory;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramImpl;
import org.processmining.models.graphbased.directed.bpmn.BPMNEdge;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.plugins.bpmnminer.causalnet.CausalNet;
import org.processmining.plugins.bpmnminer.converter.CausalNetToPetrinet;
import org.processmining.plugins.bpmnminer.fitness.RecallBPMNFitness;
import org.processmining.plugins.bpmnminer.miner.BPMNMiner;
import org.processmining.plugins.bpmnminer.types.EventLogTaskMapper;
import org.processmining.plugins.bpmnminer.types.MinerSettings;
import org.processmining.plugins.bpmnminer.ui.BPMNParameterPanel;
import org.processmining.plugins.bpmnminer.util.CausalNetBPMNDiagramConnection;
import org.processmining.plugins.log.logfilters.ui.LogFilterUI;
import org.processmining.models.graphbased.AttributeMap;

@Plugin(name = "BPMN Miner", 
		parameterLabels = {"Log", "Settings"},
		returnLabels = {"BPMN Diagram", "Miner Settings"},
		returnTypes = {
			BPMNDiagram.class,
			MinerSettings.class
		},
		userAccessible = true,
		help = "BPMN Miner Based on Heuristics")

public class BPMNMinerPlugin {
	
	@UITopiaVariant(uiLabel = "Mine BPMN model",
			affiliation = UITopiaVariant.EHV,
			author = "Seppe K.L.M. vanden Broucke",
			email = "seppe.vandenbroucke@econ.kuleuven.be",
			website = "http://econ.kuleuven.be")
	
	@PluginVariant(variantLabel = "Wizard settings", requiredParameterLabels = { 0 })
	public static Object[] mineBPMN(UIPluginContext context, XLog log) {
		MinerSettings settings = new MinerSettings();
		double nom = log.size() / ((double)log.size() + (double)settings.dependencyDivisor);
		if (nom <= 0D) nom = 0D;
		if (nom >= 0.9D) nom = 0.9D;
		settings.dependencyThreshold = nom;
		settings.l1lThreshold = nom;
		settings.l2lThreshold = nom;
		settings.suppressFitnessReport = false;
		
		BPMNParameterPanel parameters = new BPMNParameterPanel(settings);
		InteractionResult result = context.showConfiguration("Miner Parameters", parameters);
		settings = parameters.getSettings();
		
		if (result.equals(InteractionResult.CANCEL)) {
			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		if (parameters.isFilterLog()) {
			LogFilterUI filterUI = new LogFilterUI(context);
			log = filterUI.filter(log);
		}
		
		BPMNMiner miner = new BPMNMiner(log, settings);
		miner.setDeriveOriginatorSwimLanes(parameters.isMineSwimLanes());
		miner.mine();
		
		BPMNDiagram diagram = miner.getBPMNDiagram();
		if (parameters.isReportFitness())
			diagram = calculateFitness(diagram, miner.getMapper(), miner.getCausalNet(), log);
		
		 context.getProvidedObjectManager().createProvidedObject("Petri net converted from BPMN", 
				(Petrinet) CausalNetToPetrinet.convert(context, miner.getCausalNet())[0], Petrinet.class, 
				context);
		 
		 context.getConnectionManager().addConnection(
				 new CausalNetBPMNDiagramConnection("BPMN-CN connection", miner.getCausalNet(), diagram));
		
		return new Object[]{diagram, settings};
	}
	
	public static BPMNDiagram calculateFitness(BPMNDiagram originalDiagram, EventLogTaskMapper eventLogTaskMapper, CausalNet causalNet, XLog log) {
		BPMNDiagram diagram = BPMNDiagramFactory.cloneBPMNDiagram(originalDiagram);
		RecallBPMNFitness fitness = new RecallBPMNFitness(eventLogTaskMapper, causalNet);
		fitness.replayLog(log);
		
		for (int t : eventLogTaskMapper.getTasks()) {
			for (Activity a : diagram.getActivities()) {
				if (a.getLabel().equals(eventLogTaskMapper.getEventClassInteger(t))) {
					float d = fitness.getTaskInformation(t, 1);
					float n = fitness.getTaskInformation(t, 0);
					float h = (0.333333f) * (d / n);
					Color c = Color.getHSBColor(h, 1f, 1f);
					a.getAttributeMap().put(AttributeMap.FILLCOLOR, c);
				}
			}
		}
		
		System.out.println("Fitness: "+fitness.getFitness());
		
		BPMNDiagramImplP newDiagram = new BPMNDiagramImplP(diagram.getLabel()+
				" [fitness: "+fitness.getFitness()+"]");
		newDiagram.cloneFrom(diagram);
		return diagram;
	}
}

class BPMNDiagramImplP extends BPMNDiagramImpl {
	public BPMNDiagramImplP(String label) {
		super(label);
	}

	public Map<DirectedGraphElement, DirectedGraphElement> cloneFrom(
			DirectedGraph<BPMNNode, BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> graph) {
		return super.cloneFrom(graph);
	}
}
