package org.processmining.plugins.bpmnminer.visualization;

import javax.swing.JComponent;
import javax.swing.SwingConstants;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.jgraph.ProMGraphModel;
import org.processmining.models.jgraph.ProMJGraph;
import org.processmining.plugins.bpmnminer.causalnet.CausalNet;
import org.processmining.plugins.bpmnminer.converter.CausalNetToDependencyNet;
import org.processmining.plugins.bpmnminer.converter.DependencyNetToDependencyGraph;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyGraph;
import org.processmining.plugins.bpmnminer.dependencygraph.DependencyNet;
import org.processmining.plugins.bpmnminer.miner.FodinaMiner;
import org.processmining.plugins.bpmnminer.types.EventLogTaskMapper;
import org.processmining.plugins.bpmnminer.types.IntegerEventLog;
import org.processmining.plugins.bpmnminer.types.MinerSettings;
import org.processmining.plugins.bpmnminer.types.RelationalMetrics;
import org.processmining.plugins.kutoolbox.visualizators.ExportPanel;
import org.processmining.plugins.kutoolbox.visualizators.PIPPanel;
import org.processmining.plugins.kutoolbox.visualizators.ZoomPanel;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;

@Plugin(name = "Visualize and Analyze Causal Net Flows",
	parameterLabels = { "Causal Net", "Settings" }, 
	returnLabels = { "CG Flow Visualization" }, 
	returnTypes = { JComponent.class })
@Visualizer
public class FlowVisualization {

	@PluginVariant(requiredParameterLabels = { 0, 1 })
	public static JComponent visualize(PluginContext context, XLog log, MinerSettings settings) {
		EventLogTaskMapper mapper = new EventLogTaskMapper(log, settings.classifier);
		mapper.setup(settings.backwardContextSize, 
				settings.forwardContextSize, 
				settings.useUniqueStartEndTasks, 
				settings.collapseL1l,
				settings.duplicateThreshold);
		IntegerEventLog ieLog = mapper.getIntegerLog();
		FodinaMiner fodina = new FodinaMiner(ieLog, settings);
		CausalNet net = fodina.mine();
		
		// Build dependency net structure
		RelationalMetrics metrics = new RelationalMetrics(ieLog, settings);
		DependencyNet dependencyNet = CausalNetToDependencyNet.toDependencyNet(net);
		for (int t : dependencyNet.getTasks())
			for (int s : dependencyNet.getTasks())
				if ((metrics.getBasicRelations().getABMeasure(t, s) > 0) &&
						((dependencyNet.getStartTask() != s 
					&& dependencyNet.getEndTask() != t) 
						|| (s == t)))
					dependencyNet.setArc(t, s);
		DependencyGraph dependencyGraph = DependencyNetToDependencyGraph.toDependencyGraph(dependencyNet);
		ProMJGraph graph = buildJGraph(dependencyGraph);
		
		// Start view
		FlowAnalysisView view = new FlowAnalysisView(context, graph, dependencyNet, metrics, mapper, log);

		PIPPanel pip = new PIPPanel(view);
		view.addViewPanel(pip, "PIP", "View", true);

		ZoomPanel zoom = new ZoomPanel(view, DependencyGraphView.MAX_ZOOM);
		zoom.setHeight(200);
		view.addViewPanel(zoom, "Zoom", "View", true);

		ExportPanel export = new ExportPanel(view);
		view.addViewPanel(export, "Options", "Info", false);

		view.addInteractionViewports(pip, zoom);

		return view;
	}

	private static ProMJGraph buildJGraph(DependencyGraph graph) {
		ViewSpecificAttributeMap map = new ViewSpecificAttributeMap();
		GraphLayoutConnection layoutConnection = new GraphLayoutConnection(graph);
		ProMGraphModel model = new ProMGraphModel(graph);
		ProMJGraph jGraph = new ProMJGraph(model, map, layoutConnection);
		JGraphHierarchicalLayout layout = new JGraphHierarchicalLayout();
		layout.setDeterministic(false);
		layout.setCompactLayout(false);
		layout.setFineTuning(true);
		layout.setParallelEdgeSpacing(15);
		layout.setFixRoots(false);
		layout.setOrientation(map.get(graph, AttributeMap.PREF_ORIENTATION, SwingConstants.SOUTH));
		if (!layoutConnection.isLayedOut()) {
			JGraphFacade facade = new JGraphFacade(jGraph);
			facade.setOrdered(false);
			facade.setEdgePromotion(true);
			facade.setIgnoresCellsInGroups(false);
			facade.setIgnoresHiddenCells(false);
			facade.setIgnoresUnconnectedCells(false);
			facade.setDirected(true);
			facade.resetControlPoints();
			facade.run(layout, true);
			java.util.Map<?, ?> nested = facade.createNestedMap(true, true);
			jGraph.getGraphLayoutCache().edit(nested);
			layoutConnection.setLayedOut(true);
		}
		jGraph.setUpdateLayout(layout);
		layoutConnection.updated();
		return jGraph;
	}
	
	
}
